package hashproduct.popmatch;

import hashproduct.popmatch.util.*;
import hashproduct.popmatch.util.BipartiteMatcher.GED;
import static java.lang.Math.*;
import static hashproduct.popmatch.Statics.*;

public class ClassMaximalMatcher extends AbstractSolver {
	private ClassMaximalMatcher() {
		super("Class-maximal solver");
	}
	public static final ClassMaximalMatcher it = new ClassMaximalMatcher();
	
	private static boolean DEBUG = false;
	
	private static boolean shouldDeleteEdge(GED ged1, GED ged2) {
		switch (ged1) {
		case E:
			return false;
		case U:
			return ged2 == GED.O;
		case O:
			return ged2 != GED.E;
		}
		throw new AssertionError();
	}
	
	public Matching solve(Instance i) {
		BipartiteMatcher bm = new BipartiteMatcher(i.numPeople(),
			i.numPosts() + i.numPeople());
		
		// When we add a person's last resort, we set this.
		boolean[] personDone = new boolean[i.numPeople()];
		
		// True if a post has ever been in an O_i or U_i, in which case
		// we do not add new edges on it.
		boolean[] postDone = new boolean[i.numPosts()];
		
		if (DEBUG) bm.decompose(); // so we don't get nulls
		
		for (int cl = 0; ; cl++) {
			if (DEBUG) {
				System.err.println("Beginning of class " + cl + ":");
				System.err.println("Bipartite matcher OKs:");
				System.err.print("  ");
				for (int post = 0; post < i.numPosts(); post++) {
					String ged = bm.barGed[post].name();
					if (postDone[post])
						ged = ged.toLowerCase();
					System.err.print(ged + " ");
				}
				System.err.println();
				for (int person = 0; person < i.numPeople(); person++) {
					String ged = bm.fooGed[person].name();
					if (personDone[person])
						ged = ged.toLowerCase();
					System.err.print(ged + "[");
					for (int post = 0; post < i.numPosts(); post++) {
						System.err.print(
							i.rankFor(person, post) == IMPOSSIBLE_RANK ? ". "
								: bm.ok[person][post] ? i.rankFor(person, post) + " " : "? ");
					}
					int lrbar = i.numPosts() + person;
					System.err.print(bm.ok[person][lrbar]
					                               ? "LR" : "??");
					System.err.println("(" + bm.barGed[lrbar].name() + ")]");
				}
				Matching m = PressurizationMatcher.toMatching(i, bm);
				PressureAnalysis ma = Pressurizer.pressurize(m);
				System.err.println(ma.toString());
			}
			// Step 1: Add edges
			boolean somethingAdded = false;
			for (int person = 0; person < i.numPeople(); person++) {
				if (personDone[person])
					continue;
				somethingAdded = true;
				// We need to find the top-ranked available posts.
				// First, find the top rank that has available posts.
				int bestRank = LAST_RESORT_RANK;
				for (int post = 0; post < i.numPosts(); post++) {
					if (!postDone[post]) {
						int r = i.rankFor(person, post);
						bestRank = min(bestRank, r);
					}
				}
				if (bestRank == LAST_RESORT_RANK) {
					// Grab the last resort and mark the person done.
					bm.ok[person][i.numPosts() + person] = true;
					personDone[person] = true;
				} else {
					// Find all available posts of the top rank.
					for (int post = 0; post < i.numPosts(); post++)
						if (i.rankFor(person, post) == bestRank
										&& !postDone[post])
							bm.ok[person][post] = true;
				}
			}
			if (!somethingAdded)
				break; // finished; no more edges to consider
			// Step 2: Augment
			bm.augment();
			// Step 3: Decompose
			bm.decompose();
			// Step 4: Delete edges and finish posts based on the results
			// HMMM The RMM algorithm would finish people; do we?
			for (int post = 0; post < i.numPosts(); post++) {
				if (bm.barGed[post] != GED.E)
					postDone[post] = true;
			}
			for (int person = 0; person < i.numPeople(); person++) {
				if (bm.fooGed[person] != GED.E)
					personDone[person] = true;
				for (int post = 0; post < i.numPosts(); post++) {
					if (shouldDeleteEdge(bm.fooGed[person], bm.barGed[post]))
						bm.ok[person][post] = false;
				}
			}
		}
		
		return PressurizationMatcher.toMatching(i, bm);
	}
}
