package hashproduct.popmatch;
import static hashproduct.popmatch.Statics.*;

import java.util.*;

import hashproduct.popmatch.util.*;

public class Pressurizer {
	private Pressurizer() {}
	
	/* old way:
	 * pressures, constrained nonnegative via offer; highest class to person who
	 * can push the most people; empty post like pushing a lot of people via offer
	 * 
	 * new way:
	 * pressures, constrained nonnegative via offer; highest pressure to person
	 * whose knocking would allow the most to improve; check after solving that
	 * no empty post got a positive pressure
	 * 
	 * newer way (5/10): pressure per post, not per possible edge
	 * */
	
	// c and m had better be from the same instance
	public static boolean stabilizes(Pressurization c, Matching m) {
		Instance i = m.instance();
		// Strictly better posts are at strictly higher pressures,
		// and mildly better posts are at mildly higher pressures.
		for (int person = 0; person < i.numPeople(); person++) {
			int assignedPost = m.postFor(person);
			// The last resort won't be strictly better, so we can ignore it.
			for (int post = 0; post < i.numPosts(); post++)
				if (i.rankFor(person, post) < i.rankFor(person, assignedPost)) {
					if (!(c.pressureFor(post) > c.pressureFor(assignedPost)))
						return false;
				} else if (i.rankFor(person, post) == i.rankFor(person, assignedPost)) {
					if (!(c.pressureFor(post) >= c.pressureFor(assignedPost)))
						return false;
				}
		}
		// Vacant posts have zero pressure.
		for (int post = 0; post < i.numPosts(); post++)
			if (m.personFor(post) == NOBODY && c.pressureFor(post) != 0)
				return false;
		// If we get here, the classification works.
		return true;
	}
	
	public static final class PressureNode
		extends BellmanFordSolver.Node<PressureNode> {
		public final int post;
		PressureNode(int post) {
			this.post = post;
		}
	}
	
	public static PressureAnalysis pressurize(Matching m) {
		Instance i = m.instance();
		// Make our indexer.
		BellmanFordSolver<PressureNode> bfs = new BellmanFordSolver<PressureNode>();
		
		// Introduce nodes.
		PressureNode highSentinel = new PressureNode(Integer.MAX_VALUE);
		bfs.nodes.add(highSentinel);
		bfs.offer(highSentinel, 0, null);
		PressureNode[] nodes = new PressureNode[i.numPosts()];
		PressureNode[] lastResortNodes = new PressureNode[i.numPeople()];
		for (int post = 0; post < i.numPosts(); post++) {
			nodes[post] = new PressureNode(post);
			bfs.nodes.add(nodes[post]);
			bfs.offer(nodes[post], 0, null);
			nodes[post].putEdge(highSentinel, 0);
		}
		for (int person = 0; person < i.numPeople(); person++) {
			lastResortNodes[person] = new PressureNode(lastResortFor(person));
			bfs.nodes.add(lastResortNodes[person]);
			bfs.offer(lastResortNodes[person], 0, null);
			lastResortNodes[person].putEdge(highSentinel, 0);
		}
		
		// Strictly better posts are at strictly higher pressures.  Ditto mildly.
		for (int person = 0; person < i.numPeople(); person++) {
			int assignedPost = m.postFor(person);
			PressureNode apn = isALastResort(assignedPost)
				? lastResortNodes[person] : nodes[assignedPost];
			// The last resort won't be strictly better, so we can ignore it.
			for (int post = 0; post < i.numPosts(); post++)
				if (i.rankFor(person, post) < i.rankFor(person, assignedPost))
					apn.putEdge(nodes[post], -1);
				else if (i.rankFor(person, post) == i.rankFor(person, assignedPost))
					apn.putEdge(nodes[post], 0);
		}
		
		PressureNode cycled = bfs.solve();
		if (cycled == null) {
			int unpopularity = -highSentinel.dist;
			PressureNode imp = highSentinel.from;
			// Check that the vacant-post nodes all have zero pressure
			for (PressureNode n : nodes)
				if (m.personFor(n.post) == NOBODY && -n.dist != 0) {
					// Path Pareto improvement!
					unpopularity = PARETO_INEFFICIENT;
					// Restore the old tie-breaking rules.
					if (imp == highSentinel.from)
						imp = n;
				}
			// Now move all the level fields into arrays.
			int[] pressures = new int[i.numPosts()];
			for (int post = 0; post < i.numPosts(); post++)
				pressures[post] = -nodes[post].dist;
			return new PressureAnalysis(m, unpopularity,
				new PressurizationImpl(i, -highSentinel.dist, pressures),
				imp, nodes, lastResortNodes, bfs);
		} else {
			// There's a cycle--that's one way to be Pareto inefficient
			// but the BFS might give us an initial transient (grr!)
			Set<PressureNode> seen = new LinkedHashSet<PressureNode>();
			PressureNode n = cycled;
			while (!seen.contains(n)) {
				seen.add(n);
				n = n.from;
			}
			return new PressureAnalysis(m, PARETO_INEFFICIENT, null, n,
				nodes, lastResortNodes, bfs);
		}
	}
}
