package hashproduct.popmatch.threesat;
import static java.util.Arrays.*;
import static hashproduct.popmatch.threesat.Statics.*;
import static hashproduct.popmatch.Statics.*;
import hashproduct.popmatch.*;

/*
 * Note: This reduction differs from the one in the paper in that we generate one
 * more box than we need at each end of each variable chain, so that each
 * reference ``owns'' exactly one box.  This is convenient because we can then
 * allocate people and positions for all the variable references in one slab.
 */
public final class ReductionToPopMatch {
	private Instance tsi;
	private hashproduct.popmatch.Instance gpmi;
	
	public hashproduct.popmatch.threesat.Instance tsi() {
		return tsi;
	}
	public hashproduct.popmatch.Instance gpmi() {
		return gpmi;
	}
	
	// Temporary during transformation
	private int[][] ranks;
	private int[] curLinkPostPos, curLinkPostNeg;
	private int swapCurLinkPost(int ref, int newPost) {
		int[] curLinkPostArr = (isNegated(ref) ? curLinkPostNeg : curLinkPostPos);
		int var = variableOf(ref);
		int oldPost = curLinkPostArr[var];
		curLinkPostArr[var] = newPost;
		return oldPost;
	}
	
	// Fill during transformation, save for #truthValuesToMatching
	private int[][] widePosts, narrow1Posts, narrow2Posts;
	
	private static final int I = IMPOSSIBLE_RANK;
	
	private void writeRanks(int person, int postBase, int[] rs) {
		System.arraycopy(rs, 0, ranks[person], postBase, rs.length);
	}
	
	// BOXES
	private static final int internalPeoplePerBox = 3;
	private static final int pegPeoplePerBox = 3;

	private static final int widePersonOff = 0;
	private static final int boxInternalPersonOff = widePersonOff + 1;
	private static final int narrow1PersonOff = boxInternalPersonOff + internalPeoplePerBox;
	private static final int narrow2PersonOff = narrow1PersonOff + 1;
	private static final int peg2PersonOff = narrow2PersonOff + 1;
	private static final int peoplePerBox = peg2PersonOff + pegPeoplePerBox;

	private static final int internalPostsPerBox = 4;
	private static final int pegPostsPerBox = 1;
	
	private static final int boxInternalPostOff = 0;
	private static final int peg2PostOff = boxInternalPostOff + internalPostsPerBox;
	private static final int postsPerBox = peg2PostOff + pegPostsPerBox;
	
	private static final int[] boxInternalPersonRanks = {0, 1, 2, 3, I};
	private static final int[] widePersonRanks = {1, 2, 4, 3, I};
	private static final int[] narrowPersonRanks = {I, I, I, 1, I};
	private static final int[] peg2PersonRanks = {I, I, I, I, 1};
	
	private void createBox(int postOff, int personOff, int widePost,
		int narrow1Post, int narrow2Post) {
		
		int wdp = personOff + widePersonOff;
		writeRanks(wdp, postOff, widePersonRanks);
		ranks[wdp][widePost] = 0;
		
		for (int bj = 0; bj < internalPeoplePerBox; bj++)
			writeRanks(personOff + boxInternalPersonOff + bj,
				postOff, boxInternalPersonRanks);
		
		int n1p = personOff + narrow1PersonOff;
		writeRanks(n1p, postOff, narrowPersonRanks);
		ranks[n1p][narrow1Post] = 0;
		
		int n2p = personOff + narrow2PersonOff;
		writeRanks(n2p, postOff, narrowPersonRanks);
		ranks[n2p][narrow2Post] = 0;
		
		for (int bj = 0; bj < pegPeoplePerBox; bj++) {
			int wsh2p = personOff + peg2PersonOff + bj;
			writeRanks(wsh2p, postOff, peg2PersonRanks);
			ranks[wsh2p][narrow2Post] = 0;
		}
	}
	
	// POOLS
	private static final int postsPerPool = 2;
	private static final int[] poolPersonRanks = {1, 2};
	private void addPersonToPool(int person, int postOff) {
		writeRanks(person, postOff, poolPersonRanks);
	}
	
	// Allocation of people for a ref
	private static final int refBoxPersonOff = 0;
	/*
	private static final int linkBPersonOff = 0;
	private static final int boxPersonBaseOff = linkBPersonOff + 1;
	private static final int assertPersonOff = boxPersonBaseOff + peoplePerBox;
	private static final int linkFPersonOff = poolPersonOff + 1;
	*/
	private static final int refPoolPersonOff = refBoxPersonOff + peoplePerBox;
	private static final int peoplePerRef = peoplePerBox + 1;
	
	// Allocation of posts for a ref
	private static final int refBoxPostOff = 0;
	private static final int refSharedPostOff = refBoxPostOff + postsPerBox;
	private static final int refLinkFPostOff = refSharedPostOff + 1;
	private static final int postsPerRef = refLinkFPostOff + 1;
	
	// Allocation for a statement
	private static final int peoplePerStatement = 3 * peoplePerRef;
	private static final int postsPerStatement = 3 * postsPerRef + postsPerPool;
	
	private int numPeople() {
		return peoplePerStatement * tsi.numStatements();
	}
	private int numPosts() {
		return tsi.numVariables() + postsPerStatement * tsi.numStatements();
	}
	
	// Some indexing helpers
	private int variablePost(int var) {
		return var;
	}
	private int statementPostOff(int st) {
		return tsi.numVariables() + postsPerStatement * st;
	}
	private int poolPostOff(int st) {
		return statementPostOff(st);
	}
	private int refPostOff(int st, int rj) {
		return statementPostOff(st) + postsPerPool + postsPerRef * rj;
	}
	private int refPersonOff(int st, int rj) {
		return peoplePerStatement * st + peoplePerRef * rj;
	}
	
	// Create the subgraph for a reference and link it in.
	private void createRef(int st, int rj) {
		int postOff = refPostOff(st, rj), personOff = refPersonOff(st, rj);
		int ref = tsi.referenceAt(st, rj);
		
		widePosts[st][rj] = swapCurLinkPost(ref, postOff + refLinkFPostOff);
		narrow1Posts[st][rj] = postOff + refLinkFPostOff;
		// Important: narrow #2 has the peg on it.
		narrow2Posts[st][rj] = postOff + refSharedPostOff;
		createBox(postOff, personOff + refBoxPersonOff,
			widePosts[st][rj], narrow1Posts[st][rj], narrow2Posts[st][rj]);
		
		ranks[personOff + refPoolPersonOff][postOff + refSharedPostOff] = 0;
		addPersonToPool(personOff + refPoolPersonOff, poolPostOff(st));
	}
	
	// Create the entire generalized popular matching instance.
	private void createGPMI() {
		// initialize temporaries
		ranks = new int[numPeople()][numPosts()];
		for (int p = 0; p < numPeople(); p++)
			fill(ranks[p], IMPOSSIBLE_RANK);
		curLinkPostPos = new int[tsi.numVariables()];
		curLinkPostNeg = new int[tsi.numVariables()];
		for (int var = 0; var < tsi.numVariables(); var++) {
			curLinkPostPos[var] = variablePost(var);
			curLinkPostNeg[var] = variablePost(var);
		}
		widePosts = new int[tsi.numStatements()][3];
		narrow1Posts = new int[tsi.numStatements()][3];
		narrow2Posts = new int[tsi.numStatements()][3];
		
		for (int st = 0; st < tsi.numStatements(); st++)
			for (int rj = 0; rj < 3; rj++)
				createRef(st, rj);
		
		// create gpmi
		gpmi = new hashproduct.popmatch.InstanceImpl(numPeople(), numPosts(), ranks);
		
		// release temporaries
		ranks = null;
		curLinkPostPos = null;
		curLinkPostNeg = null;
	}

	public ReductionToPopMatch(Instance i) {
		this.tsi = i;
		
		createGPMI();
	}
	
	public Matching truthValuesToMatching(boolean[] truthValues) {
		if (truthValues.length != tsi.numVariables())
			throw new IllegalArgumentException();
		int[] postsFor = new int[gpmi.numPeople()];
		// Match each reference (box + pool person)
		for (int st = 0; st < tsi.numStatements(); st++) {
			int numUnsatisfiedSoFar = 0;
			for (int rj = 0; rj < 3; rj++) {
				int personOff = refPersonOff(st, rj),
					postOff = refPostOff(st, rj),
					poolPostOff = poolPostOff(st);
				int ref = tsi.referenceAt(st, rj);
				int var = variableOf(ref);
				boolean neg = isNegated(ref);
				// The internal people are always done the same way.
				for (int bj = 0; bj < 3; bj++)
					postsFor[personOff + refBoxPersonOff + boxInternalPersonOff + bj]
					         = postOff + refBoxPostOff + boxInternalPostOff + bj;
				// The peg people are always done the same way.
				int peg2pe0 = personOff + refBoxPersonOff + peg2PersonOff;
				postsFor[peg2pe0  ] = postOff + refBoxPostOff + peg2PostOff;
				postsFor[peg2pe0+1] = lastResortFor(peg2pe0+1);
				postsFor[peg2pe0+2] = lastResortFor(peg2pe0+2);
				if (truthValues[var] ^ neg) {
					// Match the box toward the center.
					// Wide person to linking post.
					postsFor[personOff + refBoxPersonOff + widePersonOff] = widePosts[st][rj];
					// *Second* narrow to the fourth post of the box.
					// We must do the narrow people this way around for the peg to protect us.
					// Sorry, this is reversed from the paper.
					postsFor[personOff + refBoxPersonOff + narrow2PersonOff] =
						postOff + refBoxPostOff + 3;
					// First to his last resort.
					postsFor[personOff + refBoxPersonOff + narrow1PersonOff] =
						lastResortFor(personOff + refBoxPersonOff + narrow2PersonOff);
					// The pool's linking person goes to the shared post.
					postsFor[personOff + refPoolPersonOff] = postOff + refSharedPostOff;
				} else {
					// Match the box away from the center.
					// Wide person to the fourth post of the box.
					postsFor[personOff + refBoxPersonOff + widePersonOff] =
						postOff + refBoxPostOff + 3;
					// Narrow people to linking posts.
					postsFor[personOff + refBoxPersonOff + narrow1PersonOff] = narrow1Posts[st][rj];
					postsFor[personOff + refBoxPersonOff + narrow2PersonOff] = narrow2Posts[st][rj];
					// The pool's linking person goes into the pool.
					postsFor[personOff + refPoolPersonOff] =
						(numUnsatisfiedSoFar == 2)
							? lastResortFor(personOff + refPoolPersonOff)
							: poolPostOff + numUnsatisfiedSoFar;
					numUnsatisfiedSoFar++;
				}
			}
		}
		
		Matching m = new MatchingImpl(gpmi, postsFor);
		// On the satisfied side of a box chain, we'll unnecessarily put the
		// unused trailing narrow person into the box.  Let that person be
		// promoted to the linking post to make our result Pareto efficient.
		m = MatchingImproverUtils.makeParetoEfficient(m);
		return m;
	}
	
	public int goodUnpopFactor() {
		return 2; // Always 2
	}
	public int goodUnpopMargin() {
		// For each reference: UM 1 from the box, UM 1 from its peg
		return tsi.numStatements() * 3 * 2;
	}
}
