package hashproduct.popmatch.util;

import java.util.*;

// Solver that supports the subset of min-cost flow that we need for LUMM.
public final class MinCostFlowSolver<N extends MinCostFlowSolver.Node<N>> {
	
	public static class Node<N extends Node<N>> {
		// Limitation: only supports one edge per pair of vertices.
		public Map<N, Edge> edges = new LinkedHashMap<N, Edge>();
		
		BFNode<N> bfNode;
	}
	
	static class BFNode<N extends Node<N>> extends BellmanFordSolver.Node<BFNode<N>> {
		N mcfNode;
		BFNode(N mcfNode) {
			this.mcfNode = mcfNode;
			mcfNode.bfNode = this;
		}
		
		// DEBUG
		@Override public String toString() {
			return mcfNode.toString();
		}
	}
	
	public static class Edge {
		// Limitation: unit capacity
		public boolean capacity;
		public int value;
		public Edge reverse;
		
		public Edge(boolean capacity, int value) {
			this.capacity = capacity;
			this.value = value;
		}
		
		// DEBUG
		@Override public String toString() {
			return (capacity ? "available" : "used") + " edge of value " + value;
		}
	}
	
	public void addEdge(N start, N end, int val) {
		Edge efwd = new Edge(true, val);
		Edge eback = new Edge(false, -val);
		efwd.reverse = eback;
		eback.reverse = efwd;
		start.edges.put(end, efwd);
		end.edges.put(start, eback);
	}
	
	// Add to me!
	public Set<N> nodes = new LinkedHashSet<N>();
	
	public int totalValue = 0;
	
	public void solve() {
		// Augment improving cycles.
		for (;;) {
			// Load our data into a BellmanFordSolver:
			BellmanFordSolver<BFNode<N>> bfs = new BellmanFordSolver<BFNode<N>>();
			// Create corresponding BFNodes.
			for (N n : nodes) {
				BFNode<N> bfn = new BFNode<N>(n); // sets n.bfNode
				bfs.nodes.add(bfn);
				// We seek a negative cycle.  To make sure we find it, add all
				// vertices as starting points.
				bfs.offer(bfn, 0, null);
			}
			// Add edges.
			for (N start : nodes) {
				for (Map.Entry<N, Edge> en : start.edges.entrySet()) {
					N end = en.getKey();
					Edge ed = en.getValue();
					if (ed.capacity)
						start.bfNode.edges.put(end.bfNode, -ed.value);
				}
			}
			// Go!
			BFNode<N> cycled = bfs.solve();
			if (cycled == null)
				// No augmenting cycle, so we're done.
				return;
			// Apply the returned cycle.
			BFNode<N> n1 = cycled;
			//Logger.out.append("Augment: " + n1);
			do {
				//Logger.out.append(" <- " + n1.from);
				Edge ed = n1.from.mcfNode.edges.get(n1.mcfNode);
				assert ed != null;
				assert ed.capacity;
				ed.capacity = false;
				assert !ed.reverse.capacity;
				ed.reverse.capacity = true;
				// Record the value gained.
				totalValue += ed.value;
				n1 = n1.from;
			} while (n1 != cycled);
			//Logger.out.append('\n');
		}
	}
}
